@@ -89,35 +89,7 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	/* JEJ - throttle sampling code */
 	priv->throttleLimit = -1;
 
-	common->wcmFlags = 0;               /* various flags */
 	common->wcmDevices = priv;
-	common->wcmProtocolLevel = WCM_PROTOCOL_4; /* protocol level */
-	common->wcmTPCButton = 0;          /* set Tablet PC button on/off */
-	common->wcmCapacity = -1;          /* Capacity is disabled */
-	common->wcmCapacityDefault = -1;    /* default to -1 when capacity isn't supported */
-					   /* 3 when capacity is supported */
-	common->wcmGestureParameters.wcmZoomDistance = 50;
-	common->wcmGestureParameters.wcmZoomDistanceDefault = 50;
-	common->wcmGestureParameters.wcmScrollDirection = 0;
-	common->wcmGestureParameters.wcmScrollDistance = 20;
-	common->wcmGestureParameters.wcmScrollDistanceDefault = 20;
-	common->wcmGestureParameters.wcmTapTime = 250;
-	common->wcmGestureParameters.wcmTapTimeDefault = 250;
-	common->wcmRotate = ROTATE_NONE;   /* default tablet rotation to off */
-	common->wcmMaxX = 0;               /* max digitizer logical X value */
-	common->wcmMaxY = 0;               /* max digitizer logical Y value */
-	common->wcmMaxTouchX = 1024;       /* max touch X value */
-	common->wcmMaxTouchY = 1024;       /* max touch Y value */
-	common->wcmMaxStripX = 4096;       /* Max fingerstrip X */
-	common->wcmMaxStripY = 4096;       /* Max fingerstrip Y */
-	common->wcmMaxtiltX = 128;	   /* Max tilt in X directory */
-	common->wcmMaxtiltY = 128;	   /* Max tilt in Y directory */
-	common->wcmCursorProxoutDistDefault = PROXOUT_INTUOS_DISTANCE; 
-			/* default to Intuos */
-	common->wcmSuppress = DEFAULT_SUPPRESS;    
-			/* transmit position if increment is superior */
-	common->wcmRawSample = DEFAULT_SAMPLES;    
-			/* number of raw data to be used to for filtering */
 
 	/* tool */
 	priv->tool = tool;
@@ -126,6 +98,9 @@ static int wcmAllocate(InputInfoPtr pInfo)
 	tool->device = pInfo;
 	/* tool->typeid is set once we know the type - see wcmSetType */
 
+	/* timers */
+	priv->serial_timer = TimerSet(NULL, 0, 0, NULL, NULL);
+
 	return 1;
 
 error:
@@ -147,6 +122,7 @@ static void wcmFree(InputInfoPtr pInfo)
 	if (!priv)
 		return;
 
+	TimerFree(priv->serial_timer);
 	free(priv->tool);
 	wcmFreeCommon(&priv->common);
 	free(priv);
@@ -249,6 +225,9 @@ static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	WacomDevicePtr *prev;
 	WacomCommonPtr common = priv->common;
 
+	if (!priv)
+		goto out;
+
 	DBG(1, priv, "\n");
 
 	/* Server 1.10 will UnInit all devices for us */
@@ -308,6 +287,7 @@ static void wcmUninit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		dev = dev->next;
 	}
 
+out:
 	wcmFree(pInfo);
 	xf86DeleteInput(pInfo, 0);
 }
@@ -436,8 +416,8 @@ static void wcmLinkTouchAndPen(InputInfoPtr pInfo)
 				if (common->wcmTouchDevice ||
 						tmpcommon->wcmTouchDevice)
 				{
-					common->tablet_type |= WCM_PENTOUCH;
-					tmpcommon->tablet_type |= WCM_PENTOUCH;
+					TabletSetFeature(common, WCM_PENTOUCH);
+					TabletSetFeature(tmpcommon, WCM_PENTOUCH);
 				}
 			}
 		}
@@ -503,7 +483,7 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	gWacomModule.wcmDrv = drv;
 
 	device = xf86SetStrOption(pInfo->options, "Device", NULL);
-	type = xf86FindOptionValue(pInfo->options, "Type");
+	type = xf86SetStrOption(pInfo->options, "Type", NULL);
 
 	/*
 	   Init process:
@@ -526,6 +506,8 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 	priv = (WacomDevicePtr) pInfo->private;
 	priv->common->device_path = device;
 	priv->name = pInfo->name;
+	priv->debugLevel = xf86SetIntOption(pInfo->options,
+					    "DebugLevel", priv->debugLevel);
 
 	/* check if the same device file has been added already */
 	if (wcmIsDuplicate(device, pInfo))
@@ -543,6 +525,8 @@ static int wcmPreInit(InputDriverPtr drv, InputInfoPtr pInfo, int flags)
 		/* initialize supported keys with the first tool on the port */
 		wcmDeviceTypeKeys(pInfo);
 
+	common->debugLevel = xf86SetIntOption(pInfo->options,
+					      "CommonDBG", common->debugLevel);
 	oldname = pInfo->name;
 
 	if (wcmIsHotpluggedDevice(pInfo))
@@ -595,7 +579,7 @@ SetupProc_fail:
 	if (common && priv)
 		common->wcmDevices = priv->next;
 
-	if (pInfo && pInfo->fd != -1)
+	if (pInfo->fd != -1)
 	{
 		close(pInfo->fd);
 		pInfo->fd = -1;
