@@ -263,7 +263,6 @@ test_suppress(void)
 	new.field = old.field;
 
 	test_above_suppress(pressure);
-	test_above_suppress(capacity);
 	test_above_suppress(throttle);
 	test_above_suppress(rotation);
 	test_above_suppress(abswheel);
@@ -522,6 +521,26 @@ static void test_set_type(void)
 #undef reset
 }
 
+static int test_flag_set(void)
+{
+	int i;
+	unsigned int flags = 0;
+
+	for (i = 0; i < sizeof(flags); i++)
+	{
+		int mask = 1 << i;
+		flags = 0;
+
+		assert(!MaskIsSet(flags, mask));
+		MaskSet(flags, mask);
+		assert(flags != 0);
+		assert(MaskIsSet(flags, mask));
+		MaskClear(flags, mask);
+		assert(!MaskIsSet(flags, mask));
+		assert(flags == 0);
+	}
+}
+
 int main(int argc, char** argv)
 {
 	test_common_ref();
@@ -532,6 +551,7 @@ int main(int argc, char** argv)
 	test_tilt_to_rotation();
 	test_mod_buttons();
 	test_set_type();
+	test_flag_set();
 	return 0;
 }
 
